# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

# pylint: disable=line-too-long, too-many-lines
"""Tests that the HTTP request generated is correct.
This does not require a cluster connection, except the test for provision application type."""

from __future__ import print_function
from os import (remove, path)
import json
import logging
from shutil import rmtree
import vcr
from msrest.authentication import Authentication
from mock import patch
from knack.testsdk import ScenarioTest
from azure.servicefabric import ServiceFabricClientAPIs
from sfctl.entry import cli
from sfctl.tests.helpers import (MOCK_CONFIG, get_mock_endpoint, set_mock_endpoint)
from sfctl.tests.mock_server import (find_localhost_free_port, start_mock_server)
from sfctl.tests.request_generation_body_validation import validate_flat_dictionary, \
    validate_create_application, validate_list_of_objects
# pylint: disable=import-error
try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote
# pylint: enable=import-error

class ServiceFabricRequestTests(ScenarioTest):
    """HTTP request generation tests for Service Fabric commands.
    This class requires a live connection to a cluster for some tests.
    This is so we generate commands in the way that the users will.
    The purpose of this test is to generate the commands and
    read the HTTP request to ensure correctness. The expected values are hard-coded.
    The VCR library records all HTTP requests into
    a file. For the sake of clarity, each test writes to its own file.
    The tests should then read the file to validate correctness. For on the fly debugging,
    printing to stdout does not print to the terminal/command line.
    Please use other outputs, such as stderr."""

    def __init__(self, method_name):
        cli_env = cli()
        super(ServiceFabricRequestTests, self).__init__(cli_env, method_name)

        # Save the value of cluster endpoint set by the user
        self.old_endpoint = get_mock_endpoint()
        self.port = find_localhost_free_port()

        # Start mock server
        start_mock_server(self.port)

        # Delete the generated vcr files before running the tests
        # These files are generated by knack (ScenarioTest),
        # which we are not using since we generate our
        # own vcr files in JSON rather than yaml.
        try:
            rmtree(path.join(path.dirname(__file__), 'recordings'))
        except OSError:
            # if the folder doesn't exist, then there's nothing for us to do here
            pass

    def __exit__(self, exception_type, exception_value, traceback):
        # Revert the environment variables we changed during the test to what the user set.
        set_mock_endpoint(self.old_endpoint)

    @patch('sfctl.config.CLIConfig', new=MOCK_CONFIG)
    def validate_command_succeeds(self, command):
        """
        Validate that the given command runs and returns with success
        :param command: str - see command parameter for validate_command method
        :return: None
        """
        try:
            self.cmd(command)
        except BaseException as exception:  # pylint: disable=broad-except
            self.fail(
                'ERROR while running command "{0}". Error: "{1}"'.format(command, str(exception)))

    @patch('sfctl.config.CLIConfig', new=MOCK_CONFIG)
    def validate_command(self, command, method, url_path, query, body=None,  # pylint: disable=too-many-locals, too-many-arguments
                         body_verifier=None, command_as_func=False, command_args=None):
        """
        This method takes the command passed in and runs the sfctl command specified.
        It records the input and output in a text file. It then validates the contexts of
        the text file against the provided expected results.
        All provided parameters other than self and command are expected values, rather than
        actual values by running the sfctl command.

        Some considerations:
            - This test doesn't currently test for things like different combinations of
              input parameters.
            - Ordering of the URI isn't enforced
              (path can appear before query for example)
            - API version is tested as a query parameter right now.
              Optimizations can be considered later.

        command (str): The command as it will appear in command line.
                For example, if a command if sfctl application list in
                command line, then use string 'application list'.
        method (str): The method of the HTTP request.
                For example, GET, POST, PUT, etc.
        body (str or dict): The body as a string. Set value to None if
                no validation is required. What format this string should be in
                depends on the body_verifier function which is passed in. If a
                function is not passed, but this value is not None, then a simple
                string comparison will be done.
                If the format is a str, it will be turned into an object via json.loads. Otherwise,
                an object/dict is expected and no conversion will be done. It will be passed
                directly to the body verifier in that case.
        body_verifier (function): A function which returns true if the passed
                in body (str) passes validation. The method should take:
                    1. Command being run as a string (e.g. 'cluster report-health'),
                    2. Actual, and the expected bodies as its inputs, in that order.
                       The actual and expected bodies will be passed in as objects
                       created by calling json.loads.
                Put None to skip this validation.
                If this is set to None but a body string is provided,
                validation will occur against the string as a string
                comparison.
        path (str): The section of the URI after host name and before the
                query portion. Do not include the port number.
        query (str list): A list of strings representing the query portion
                of the URI. Follows the format of ["parameter_name=parameter_value", ...].
                Make sure to include the API version of the HTTP request if
                required.
        command_as_func (bool): Indicates that the command parameter passed in is a function
                rather than a string. Call the function directly, with optional command_args.
        command_args (kwargs): Applicable only if command_as_func is True.
                Dictionary of name value pairs to be passed in a arguments
                to the command function.
        """

        # For testing purposes, write to stderr with color

        generated_file_path = 'paths_generation_test.json'

        # In case this file was created and not deleted by a previous test,
        # delete it here
        try:
            remove(generated_file_path)
        except OSError:
            # if the file doesn't exist, then there's nothing for us to do here
            pass

        # This calls the command and the HTTP request it recorded into
        # generated_file_path

        # Reduce noise in test output for this test only
        logging.disable(logging.INFO)
        with vcr.use_cassette(generated_file_path, record_mode='all', serializer='json'):
            try:
                if not command_as_func:
                    self.cmd(command)
                else:
                    command(command_args)
            except BaseException as exception:  # pylint: disable=bare-except
                self.fail('ERROR while running command "{0}". Error: "{1}"'.format(command, str(exception)))

        # re-enable logging
        logging.disable(logging.NOTSET)

        # Read recorded JSON file
        with open(generated_file_path, 'r') as http_recording_file:
            json_str = http_recording_file.read()
            vcr_recording = json.loads(json_str)

            # The responses create an array of request and other objects.
            # the numbers (for indexing) represent which request was made
            # first. The ordering is determined by the ordering of calls to self.cmd.
            # see outputted JSON file at generated_file_path for more details.
            recording = vcr_recording['interactions'][0]['request']

            # Validate method (GET, POST, etc)
            recording_method = recording['method']

            self.assertEqual(method, recording_method,
                             msg='Method mismatch. Expected {0}. Got {1}'.format(recording_method, method))

            # Validate body
            recording_body = recording['body']

            # body here is the expected body
            # recording_body is the actual value
            if body is not None and body_verifier is not None:
                actual_obj = json.loads(recording_body)
                expected_obj = None
                if isinstance(body, str):
                    expected_obj = json.loads(body)
                else:
                    expected_obj = body
                self.assertTrue(body_verifier(command, actual_obj, expected_obj),
                                msg='body_verifier command failed. The command should either print error'
                                    'messages, or call its own asserts.')
            elif (body is not None) and (body_verifier is None):
                self.assertEqual(recording_body, body,
                                 msg='Performing a string comparison for body.'
                                     'Expected "{0}". Got "{1}"'.format(recording_body, body))

            # Get HTTP URI
            recording_uri = recording['uri']
            # An example URI
            # 'http://url:port/AppTypes/$/Provision?api-version=6.1'

            # Check that path is in recording_uri
            uri_path_start = 'http://localhost:' + str(self.port) + url_path
            self.assertIn(uri_path_start, recording_uri,
                          msg='Uri starting portion "{0}" was not found in generated URI "{1}"'.format(uri_path_start, recording_uri))

            # Check that the string appears in recording_uri
            for string in query:
                self.assertTrue((string in recording_uri),
                                msg='"{0}" not found in URI query section "{1}"'.format(string, recording_uri))

        # If this test reaches here, then this test is successful.
        remove(generated_file_path)  # Clean up

    def test_paths_generation(self):
        """ This test calls all commands that exist within sfctl.
        The commands are routed to a mock cluster which always returns
        success.We then read the values of the URL and other request
        features to determine that the command is working as expected
        (generating the correct URL). """

        # Set test URL path to that of our mock server
        set_mock_endpoint('http://localhost:' + str(self.port))

        # Call test
        self.paths_generation_helper()

    def paths_generation_helper(self):  # pylint: disable=too-many-statements, too-many-locals
        """ Lists all the commands to be tested and their expected values.
        Expected values here refer to the expected URI that is generated
        and sent to the cluster."""

        sample_path_base = '@' + path.join(path.dirname(__file__), 'sample_json')

        # Get current directory
        dir_path = path.dirname(path.realpath(__file__))
        dir_path_sample_json = path.join(dir_path, 'sample_json')
        app_capacity_metrics = path.join(dir_path_sample_json, 'sample_application_capacity_metric_descriptions.txt')
        app_parameters = path.join(dir_path_sample_json, 'sample_application_parameters.txt')

        # While in commandline, you can pass in something like --metrics=C:/SomeFolder/file.txt,
        # in the strings below, we need to escape the "\"
        app_capacity_metrics = app_capacity_metrics.replace('\\', '\\\\')
        app_parameters = app_parameters.replace('\\', '\\\\')

        dir_path_one_file = path.join(dir_path, 'one_file')
        dir_path_one_file = dir_path_one_file.replace('\\', '\\\\')

        # The commands which don't affect or query the cluster
        # Specifically, cluster select and show-connection
        self.validate_command_succeeds('cluster select --endpoint=' + get_mock_endpoint())
        self.validate_command_succeeds('cluster show-connection')
        # This test (app upload) will no longer work in paths generation because I've put it in a
        # separate process,
        # so it's unable to be caught by vcr. Move this to just checking that it runs without
        # crashing.
        self.validate_command_succeeds('application upload --path=' + dir_path_one_file)

        # Settings
        # Note: always run the -off command last so that tests don't crowd our telemetry
        # Before testing starts, telemetry is turned off
        self.validate_command_succeeds('sfctl settings telemetry set-telemetry --on')
        self.validate_command_succeeds('sfctl settings telemetry set-telemetry --off')

        # Application Type Commands
        self.validate_command(  # provision-application-type image-store
            'application provision --application-type-build-path=test_path',
            'POST',
            '/ApplicationTypes/$/Provision',
            ['api-version=6.2', 'timeout=60'],
            ('{"Kind": "ImageStorePath", '
             '"Async": false, '
             '"ApplicationTypeBuildPath": "test_path"}'),
            validate_flat_dictionary)

        self.validate_command(  # provision-application-type external-store
            'application provision --external-provision '
            '--application-package-download-uri=test_path --application-type-name=name '
            '--application-type-version=version',
            'POST',
            '/ApplicationTypes/$/Provision',
            ['api-version=6.2', 'timeout=60'],
            ('{"Kind": "ExternalStore", '
             '"Async": false, '
             '"ApplicationPackageDownloadUri": "test_path", '
             '"ApplicationTypeName": "name", '
             '"ApplicationTypeVersion": "version"}'),
            validate_flat_dictionary)

        # Cluster commands:
        # Select command tested elsewhere.
        # Cluster upgrade and upgrade-update test to be added later
        self.validate_command(  # get-cluster-version
            ServiceFabricClientAPIs.get_cluster_version,
            'GET',
            '/$/GetClusterVersion',
            ['api-version=6.4'],
            command_as_func=True,
            command_args=ServiceFabricClientAPIs(credentials=Authentication(),
                                                 base_url=get_mock_endpoint()))
        self.validate_command(   # config-versions
            'sfctl cluster config-versions --config-version=version',
            'GET',
            '/$/GetProvisionedConfigVersions',
            ['api-version=6.0', 'ConfigVersion=version'])
        self.validate_command(  # health - Use exclude-health-statistics param
            'sfctl cluster health --applications-health-state-filter=2 '
            '--events-health-state-filter=2 --exclude-health-statistics '
            '--nodes-health-state-filter=2',
            'GET',
            '/$/GetClusterHealth',
            ['api-version=6.0',
             'NodesHealthStateFilter=2',
             'ApplicationsHealthStateFilter=2',
             'EventsHealthStateFilter=2',
             'ExcludeHealthStatistics=true'])
        # Use include-system-application-health-statistics param
        self.validate_command(  # health
            'sfctl cluster health --applications-health-state-filter=2 '
            '--events-health-state-filter=2 '
            '--include-system-application-health-statistics '
            '--nodes-health-state-filter=2',
            'GET',
            '/$/GetClusterHealth',
            ['api-version=6.0',
             'NodesHealthStateFilter=2',
             'ApplicationsHealthStateFilter=2',
             'EventsHealthStateFilter=2',
             'IncludeSystemApplicationHealthStatistics=true'])
        self.validate_command(  # manifest
            'sfctl cluster manifest',
            'GET',
            '/$/GetClusterManifest',
            ['api-version=6.0'])
        self.validate_command(  # provision
            'sfctl cluster provision --cluster-manifest-file-path=value --code-file-path=value2',
            'POST',
            '/$/Provision',
            ['api-version=6.0'],
            '{"CodeFilePath":"value2","ClusterManifestFilePath":"value"}',
            validate_flat_dictionary)
        # "P3Y6M4DT12H30M5S" represents a duration of "three years,
        # six months, four days, twelve hours, thirty minutes,
        # and five seconds".
        self.validate_command(  # report-health
            'sfctl cluster report-health --immediate --source-id=ID' +
            ' --health-property=Property ' +
            '--health-state=Warning --ttl=P3Y6M4DT12H30M5S ' +
            '--description=Description ' +
            '--sequence-number=10 --remove-when-expired',
            'POST',
            '/$/ReportClusterHealth',
            ['api-version=6.0', 'Immediate=true'],
            ('{"SourceId": "ID", '
             '"Property": "Property", '
             '"HealthState": "Warning", '
             '"TimeToLiveInMilliSeconds": "P3Y6M4DT12H30M5S", '
             '"Description": "Description", '
             '"SequenceNumber": "10", '
             '"RemoveWhenExpired": true}'),
            validate_flat_dictionary)
        self.validate_command(  # unprovision
            'sfctl cluster unprovision --code-version=code --config-version=config',
            'POST',
            '/$/Unprovision',
            ['api-version=6.0'],
            '{"CodeVersion":"code","ConfigVersion":"config"}',
            validate_flat_dictionary)
        self.validate_command(  # upgrade-resume
            'sfctl cluster upgrade-resume --upgrade-domain=UD2',
            'POST',
            '/$/MoveToNextUpgradeDomain',
            ['api-version=6.0'],
            '{"UpgradeDomain":"UD2"}',
            validate_flat_dictionary)
        self.validate_command(  # upgrade-rollback
            'sfctl cluster upgrade-rollback',
            'POST',
            '/$/RollbackUpgrade',
            ['api-version=6.0'])
        self.validate_command(  # upgrade-status
            'sfctl cluster upgrade-status',
            'GET',
            '/$/GetUpgradeProgress',
            ['api-version=6.0'])

        # Node Commands:
        self.validate_command(  # disable
            'sfctl node disable --node-name=nodeName --deactivation-intent=Pause',
            'POST',
            '/Nodes/nodeName/$/Deactivate',
            ['api-version=6.0'],
            '{"DeactivationIntent":"Pause"}',
            validate_flat_dictionary)
        self.validate_command(  # enable
            'sfctl node enable --node-name=nodeName',
            'POST',
            '/Nodes/nodeName/$/Activate',
            ['api-version=6.0'])
        self.validate_command(  # health
            'sfctl node health --node-name=nodeName --events-health-state-filter=2',
            'GET',
            '/Nodes/nodeName/$/GetHealth',
            ['api-version=6.0', 'EventsHealthStateFilter=2'])
        self.validate_command(  # info
            'sfctl node info --node-name=nodeName',
            'GET',
            '/Nodes/nodeName',
            ['api-version=6.0'])
        self.validate_command(  # list
            'sfctl node list --continuation-token=nodeId --node-status-filter=up --max-results=3',
            'GET',
            '/Nodes',
            ['api-version=6.3', 'ContinuationToken=nodeId', 'NodeStatusFilter=up', 'MaxResults=3'])
        self.validate_command(  # load
            'sfctl node load --node-name=nodeName',
            'GET',
            '/Nodes/nodeName/$/GetLoadInformation',
            ['api-version=6.0'])
        self.validate_command(  # remove-state
            'sfctl node remove-state --node-name=nodeName',
            'POST',
            '/Nodes/nodeName/$/RemoveNodeState',
            ['api-version=6.0'])
        self.validate_command(  # report-health
            'sfctl node report-health --node-name=nodeName --immediate --source-id=ID --health-property=Property ' +
            '--health-state=Warning --ttl=P3Y6M4DT12H30M5S --description=Description ' +
            '--sequence-number=10 --remove-when-expired=true',
            'POST',
            '/Nodes/nodeName/$/ReportHealth',
            ['api-version=6.0', 'Immediate=true'],
            ('{"SourceId": "ID", '
             '"Property": "Property", '
             '"HealthState": "Warning", '
             '"TimeToLiveInMilliSeconds": "P3Y6M4DT12H30M5S", '
             '"Description": "Description", '
             '"SequenceNumber": "10", '
             '"RemoveWhenExpired": true}'),
            validate_flat_dictionary)
        self.validate_command(  # restart
            'sfctl node restart --node-name=nodeName --node-instance-id=ID --create-fabric-dump=True',
            'POST',
            '/Nodes/nodeName/$/Restart',
            ['api-version=6.0'],
            '{"CreateFabricDump":"True", "NodeInstanceId":"ID"}',
            validate_flat_dictionary)
        self.validate_command(  # remove-configuration-overrides
            'sfctl node remove-configuration-overrides --node-name=nodeName',
            'DELETE',
            '/Nodes/nodeName/$/RemoveConfigurationOverrides',
            ['api-version=7.0'])
        self.validate_command(  # get-configuration-overrides
            'sfctl node get-configuration-overrides --node-name=nodeName',
            'GET',
            '/Nodes/nodeName/$/GetConfigurationOverrides',
            ['api-version=7.0'])
        configuration_override_parameters = path.join(sample_path_base, 'sample_config_override.txt').replace('/', '//').replace('\\', '\\\\')
        self.validate_command(  # add-configuration-parameter-overrides
            ('sfctl node add-configuration-parameter-overrides --node-name=nodeName '
             '--config-parameter-override-list={0}').format(configuration_override_parameters),
            'POST',
            '/Nodes/nodeName/$/AddConfigurationParameterOverrides',
            ['api-version=7.0'],
            ('[{"SectionName": "PlacementAndLoadBalancing", '
             '"ParameterName": "DummyPLBEnabled", '
             '"ParameterValue": "False"}]'),
            validate_list_of_objects)

        tags = "tagA,tagB,tagC"
        self.validate_command(  # add-node-tags
            'sfctl node add-node-tags --node-name=nodeName --tags=' + tags,
            'POST',
            '/Nodes/nodeName/$/AddNodeTags',
            ['api-version=7.2'],
            ('["tagA", "tagB", "tagC"]'),
            validate_list_of_objects)
        self.validate_command(  # remove-node-tags
            'sfctl node remove-node-tags --node-name=nodeName --tags=' + tags,
            'POST',
            '/Nodes/nodeName/$/RemoveNodeTags',
            ['api-version=7.0'],
            ('["tagA", "tagB", "tagC"]'),
            validate_list_of_objects)

        # container commands
        self.validate_command(  # get container logs
            'sfctl container invoke-api --node-name Node01 --application-id samples/winnodejs '
            '--service-manifest-name NodeServicePackage --code-package-name NodeService.Code '
            '--code-package-instance-id 131668159770315380 --container-api-uri-path "/containers/{id}/logs?stdout=true&stderr=true"',
            'POST',
            '/Nodes/Node01/$/GetApplications/samples/winnodejs/$/GetCodePackages/$/ContainerApi',
            ['api-version=6.2', 'ServiceManifestName=NodeServicePackage', 'CodePackageName=NodeService.Code', 'CodePackageInstanceId=131668159770315380', 'timeout=60'],
            '{"UriPath": "/containers/{id}/logs?stdout=true&stderr=true"}',
            validate_flat_dictionary)
        self.validate_command(  # get container logs
            'sfctl container logs --node-name Node01 --application-id samples/winnodejs '
            '--service-manifest-name NodeServicePackage --code-package-name NodeService.Code '
            '--code-package-instance-id 131668159770315380',
            'POST',
            '/Nodes/Node01/$/GetApplications/samples/winnodejs/$/GetCodePackages/$/ContainerApi',
            ['api-version=6.2', 'ServiceManifestName=NodeServicePackage', 'CodePackageName=NodeService.Code', 'CodePackageInstanceId=131668159770315380', 'timeout=60'],
            '{"UriPath": "/containers/{id}/logs?stdout=true&stderr=true"}',
            validate_flat_dictionary)
        self.validate_command(  # update container
            'sfctl container invoke-api --node-name N0020 --application-id nodejs1 --service-manifest-name NodeOnSF '
            '--code-package-name Code --code-package-instance-id 131673596679688285 --container-api-uri-path "/containers/{id}/update"'
            ' --container-api-http-verb=POST --container-api-body "DummyRequestBody"',  # Manual testing with a JSON string for "--container-api-body" works,
            # Have to pass "DummyRequestBody" here since a real JSON string confuses test validation code.
            'POST',
            '/Nodes/N0020/$/GetApplications/nodejs1/$/GetCodePackages/$/ContainerApi',
            ['api-version=6.2', 'ServiceManifestName=NodeOnSF', 'CodePackageName=Code', 'CodePackageInstanceId=131673596679688285', 'timeout=60'],
            ('{"UriPath": "/containers/{id}/update", '
             '"HttpVerb": "POST", '
             '"Body": "DummyRequestBody"}'),
            validate_flat_dictionary)

        # Application Commands:
        # Application upgrade  and create is not tested for all parameters
        # application upload tested as part of a custom command as well

        self.validate_command(  # create
            'application create --app-name=fabric:/application1 --app-type=applicationType --app-version=1 --max-node-count=3 --min-node-count=2 '
            '--metrics=@' + app_capacity_metrics +
            ' --parameters=@' + app_parameters,
            'POST',
            '/Applications/$/Create',
            ['api-version=6.0'],
            {
                'Name': 'fabric:/application1',
                'TypeName': 'applicationType',
                'TypeVersion': '1',
                'ParameterList': [[('Key', 'Key'), ('Value', 'Value')]],
                'MinNodes': 2,
                'MaxNodes': 3,
                'ApplicationMetricDescriptions': [[('MaximumCapacity', 5), ('Name', 'some_name'),
                                                   ('ReservationCapacity', 3),
                                                   ('TotalApplicationCapacity', 8)]]
            },
            validate_create_application)

        # Running this test with more than one file in the folder will cause some JSON
        # deserialization errors.
        self.validate_command(  # delete
            'application delete --application-id=application~Id --force-remove=true',
            'POST',
            '/Applications/application~Id/$/Delete',
            ['api-version=6.0', 'ForceRemove=true'])
        self.validate_command(  # deployed
            'application deployed --application-id=application~Id --node-name=nodeName --include-health-state',
            'GET',
            '/Nodes/nodeName/$/GetApplications/application~Id',
            ['api-version=6.1', 'IncludeHealthState=true'])
        self.validate_command(  # deployed-list
            'application deployed-list --node-name=nodeName --continuation-token=token --include-health-state --max-results=10',
            'GET',
            '/Nodes/nodeName/$/GetApplications',
            ['api-version=6.1', 'ContinuationToken=token', 'IncludeHealthState=true', 'MaxResults=10'])
        self.validate_command(  # deployed-health
            'application deployed-health --node-name=nodeName --application-id=app~id --deployed-service-packages-health-state-filter=2 ' +
            '--events-health-state-filter=2 --exclude-health-statistics',
            'GET',
            '/Nodes/nodeName/$/GetApplications/app~id/$/GetHealth',
            ['api-version=6.0', 'EventsHealthStateFilter=2', 'ExcludeHealthStatistics=true', 'DeployedServicePackagesHealthStateFilter=2'])
        self.validate_command(  # health
            'application health --application-id=app~id --deployed-applications-health-state-filter=2 --services-health-state-filter=2 ' +
            '--events-health-state-filter=2 --exclude-health-statistics',
            'GET',
            '/Applications/app~id/$/GetHealth',
            ['api-version=6.0', 'EventsHealthStateFilter=2', 'ExcludeHealthStatistics=true', 'DeployedApplicationsHealthStateFilter=2',
             'ServicesHealthStateFilter=2'])
        self.validate_command(  # info
            'application info --application-id=application~Id --exclude-application-parameters',
            'GET',
            '/Applications/application~Id',
            ['api-version=6.0', 'ExcludeApplicationParameters=true'])
        self.validate_command(  # list
            'application list --application-type-name=name --continuation-token=token --exclude-application-parameters --max-results=10',
            'GET',
            '/Applications',
            ['api-version=6.1', 'ContinuationToken=token', 'ExcludeApplicationParameters=true', 'MaxResults=10', 'ApplicationTypeName=name'])
        self.validate_command(  # load
            'application load --application-id=application~Id',
            'GET',
            '/Applications/application~Id/$/GetLoadInformation',
            ['api-version=6.0'])
        self.validate_command(  # report-health
            'application report-health --application-id=id --immediate --source-id=ID --health-property=Property ' +
            '--health-state=Warning --ttl=P3Y6M4DT12H30M5S --description=Description ' +
            '--sequence-number=10 --remove-when-expired=true',
            'POST',
            '/Applications/id/$/ReportHealth',
            ['api-version=6.0', 'Immediate=true'],
            ('{"SourceId": "ID", '
             '"Property": "Property", '
             '"HealthState": "Warning", '
             '"TimeToLiveInMilliSeconds": "P3Y6M4DT12H30M5S", '
             '"Description": "Description", '
             '"SequenceNumber": "10", '
             '"RemoveWhenExpired": true}'),
            validate_flat_dictionary)

        app_params = path.join(sample_path_base, 'sample_application_parameters.txt').replace('/', '//').replace('\\', '\\\\')
        default_service_type_health_policy = path.join(sample_path_base, 'sample_default_service_type_health_policy.txt').replace('/', '//').replace('\\', '\\\\')  # pylint: disable=invalid-name
        service_type_health_policy_map = path.join(sample_path_base, 'sample_service_type_health_policy_map.txt').replace('/', '//').replace('\\', '\\\\')
        sample_application_capacity_metric_descriptions = path.join(sample_path_base, 'sample_application_capacity_metric_descriptions.txt').replace('/', '//').replace('\\', '\\\\')  # pylint: disable=invalid-name
        sample_application_parameters = path.join(sample_path_base, 'sample_application_parameters.txt').replace('/', '//').replace('\\', '\\\\')

        # application upgrade and application create does not currently test the body
        # Add this in later.
        self.validate_command(  # upgrade - not all parameters tested
            ('application upgrade '
             '--application-id=app '
             '--application-version=1.0.0 '
             '--parameters={0} '
             '--default-service-health-policy={1} '
             '--failure-action=Rollback '
             '--force-restart=true '
             '--health-check-retry-timeout=PT0H21M0S '
             '--health-check-stable-duration=PT0H21M0S '
             '--health-check-wait-duration=PT0H21M0S '
             '--max-unhealthy-apps=20 '
             '--mode=Monitored '
             '--replica-set-check-timeout=10 '
             '--service-health-policy={2} '
             '--upgrade-domain-timeout=some_timeout '
             '--upgrade-timeout=some_timeout2 '
             '--warning-as-error').format(
                 app_params, default_service_type_health_policy, service_type_health_policy_map),
            'POST',
            '/Applications/app/$/Upgrade',
            ['api-version=6.0'])

        self.validate_command(  # create
            ('application create '
             '--app-name=fabric:/app '
             '--app-type=test-type '
             '--app-version=1.0.0 '
             '--max-node-count=3 '
             '--min-node-count=1 '
             '--metrics={0} '
             '--parameters={1} ').format(sample_application_capacity_metric_descriptions, sample_application_parameters),
            'POST',
            '/Applications/$/Create',
            ['api-version=6.0'])

        self.validate_command(  # upgrade-resume
            'application upgrade-resume --application-id=application~Id --upgrade-domain-name=UD2',
            'POST',
            '/Applications/application~Id/$/MoveToNextUpgradeDomain',
            ['api-version=6.0'],
            '{"UpgradeDomainName":"UD2"}',
            validate_flat_dictionary)
        self.validate_command(  # upgrade-rollback
            'application upgrade-rollback --application-id=application~Id',
            'POST',
            '/Applications/application~Id/$/RollbackUpgrade',
            ['api-version=6.0'])
        self.validate_command(  # upgrade-status
            'application upgrade-status --application-id=application~Id',
            'GET',
            '/Applications/application~Id/$/GetUpgradeProgress',
            ['api-version=6.0'])

        # Compose
        # create is not tested here because it requires reading in a file.
        self.validate_command(  # list
            'compose list --continuation-token=token --max-results=10',
            'GET',
            '/ComposeDeployments',
            ['api-version=6.0-preview', 'ContinuationToken=token', 'MaxResults=10'])
        self.validate_command(  # status
            'compose status --deployment-name=deploymentName',
            'GET',
            '/ComposeDeployments/deploymentName',
            ['api-version=6.0-preview'])
        self.validate_command(  # remove
            'compose remove --deployment-name=deploymentName',
            'POST',
            '/ComposeDeployments/deploymentName/$/Delete',
            ['api-version=6.0-preview'])
        self.validate_command(  # upgrade-status
            'compose upgrade-status --deployment-name=deploymentName',
            'GET',
            '/ComposeDeployments/deploymentName/$/GetUpgradeProgress',
            ['api-version=6.0-preview'])
        self.validate_command(  # upgrade-rollback
            'compose upgrade-rollback --deployment-name=deploymentName',
            'POST',
            '/ComposeDeployments/deploymentName/$/RollbackUpgrade',
            ['api-version=6.4-preview'])

        # IS:
        self.validate_command(  # command
            'is command --command=cmd --service-id=id',
            'POST',
            '/$/InvokeInfrastructureCommand',
            ['api-version=6.0', 'Command=cmd', 'ServiceId=id'])
        self.validate_command(  # query
            'is query --command=cmd --service-id=id',
            'GET',
            '/$/InvokeInfrastructureQuery',
            ['api-version=6.0', 'Command=cmd', 'ServiceId=id'])

        # Partition
        self.validate_command(  # health
            'partition health --partition-id=id --events-health-state-filter=2 --exclude-health-statistics --replicas-health-state-filter=2',
            'GET',
            '/Partitions/id/$/GetHealth',
            ['api-version=6.0', 'EventsHealthStateFilter=2', 'ReplicasHealthStateFilter=2', 'ExcludeHealthStatistics=true'])
        self.validate_command(  # info
            'partition info --partition-id=id',
            'GET',
            '/Partitions/id',
            ['api-version=6.0'])
        self.validate_command(  # list
            'partition list --service-id=fabric:/app/id --continuation-token=ct',
            'GET',
            '/Services/fabric:/app/id/$/GetPartitions',
            ['api-version=6.4', 'ContinuationToken=ct'])
        self.validate_command(  # load
            'partition load --partition-id=id',
            'GET',
            '/Partitions/id/$/GetLoadInformation',
            ['api-version=6.0'])
        self.validate_command(  # load-reset
            'partition load-reset --partition-id=id',
            'POST',
            '/Partitions/id/$/ResetLoad',
            ['api-version=6.0'])
        self.validate_command(  # recover
            'partition recover --partition-id=id',
            'POST',
            '/Partitions/id/$/Recover',
            ['api-version=6.0'])
        self.validate_command(  # recover-all
            'partition recover-all',
            'POST',
            '/$/RecoverAllPartitions',
            ['api-version=6.0'])
        self.validate_command(  # svc-name
            'partition svc-name --partition-id=id',
            'GET',
            '/Partitions/id/$/GetServiceName',
            ['api-version=6.0'])
        self.validate_command(  # report-health
            'partition report-health --partition-id=id --immediate --source-id=ID --health-property=Property ' +
            '--health-state=Warning --ttl=P3Y6M4DT12H30M5S --description=Description ' +
            '--sequence-number=10 --remove-when-expired=true',
            'POST',
            '/Partitions/id/$/ReportHealth',
            ['api-version=6.0', 'Immediate=true'],
            ('{"SourceId": "ID", '
             '"Property": "Property", '
             '"HealthState": "Warning", '
             '"TimeToLiveInMilliSeconds": "P3Y6M4DT12H30M5S", '
             '"Description": "Description", '
             '"SequenceNumber": "10", '
             '"RemoveWhenExpired": true}'),
            validate_flat_dictionary)
        self.validate_command(  # move-primary-replica
            'partition move-primary-replica --partition-id=id --node-name=nodeName --ignore-constraints',
            'POST',
            '/Partitions/id/$/MovePrimaryReplica',
            ['api-version=6.5', 'IgnoreConstraints=true', 'timeout=60'])
        self.validate_command(  # move-secondary-replica
            'partition move-secondary-replica --partition-id=id --current-node-name=nodeName --new-node-name=nodeName --ignore-constraints',
            'POST',
            '/Partitions/id/$/MoveSecondaryReplica',
            ['api-version=6.5', 'IgnoreConstraints=true', 'timeout=60'])
        self.validate_command( # loaded-partitions
           'partition get-loaded-partition-info-list --metric-name=Count',
           'GET',
           '/$/GetLoadedPartitionInfoList',
           ['api-version=8.0', 'MetricName=Count'])
        self.validate_command(  # move-instance
            'partition move-instance --service-id=fabric:/app/id --partition-id=id --current-node-name=nodeName --new-node-name=nodeName --ignore-constraints',
            'POST',
            '/Services/fabric:/app/id/$/GetPartitions/id/$/MoveInstance',
            ['api-version=8.0', 'IgnoreConstraints=true', 'timeout=60'])
        # Property
        value = '"{\\"Kind\\": \\"String\\", \\"Data\\": \\"data\\"}"'
        self.validate_command(  # put
            'property put --name-id=name --property-name=property --custom-id-type=type --value=' + value,
            'PUT',
            '/Names/name/$/GetProperty',
            ['api-version=6.0'],
            '{"PropertyName":"property", "CustomTypeId":"type", "Value":{"Kind": "String", "Data": "data"}}',
            validate_flat_dictionary)
        self.validate_command(  # get
            'property get --name-id=name --property-name=property',
            'GET',
            '/Names/name/$/GetProperty',
            ['api-version=6.0', 'PropertyName=property'])
        self.validate_command(  # list
            'property list --name-id=name --continuation-token=ct --include-values',
            'GET',
            '/Names/name/$/GetProperties',
            ['api-version=6.0', 'IncludeValues=true', 'ContinuationToken=ct'])
        self.validate_command(  # delete
            'property delete --name-id=name --property-name=property',
            'DELETE',
            '/Names/name/$/GetProperty',
            ['api-version=6.0', 'PropertyName=property'])

        # Replica
        self.validate_command(  # deployed
            'replica deployed --node-name=nodeName --partition-id=id --replica-id=replicaId',
            'GET',
            '/Nodes/nodeName/$/GetPartitions/id/$/GetReplicas/replicaId/$/GetDetail',
            ['api-version=6.0'])
        self.validate_command(  # deployed-list
            'replica deployed-list --node-name=nodeName --partition-id=id --application-id=applicationId --service-manifest-name=serviceManifestName',
            'GET',
            '/Nodes/nodeName/$/GetApplications/applicationId/$/GetReplicas',
            ['api-version=6.0', 'PartitionId=id', 'ServiceManifestName=serviceManifestName'])
        self.validate_command(  # health
            'replica health --events-health-state-filter=2 --partition-id=id --replica-id=replicaId',
            'GET',
            '/Partitions/id/$/GetReplicas/replicaId/$/GetHealth',
            ['api-version=6.0', 'EventsHealthStateFilter=2'])
        self.validate_command(  # info
            'replica info --partition-id=id --replica-id=replicaId',
            'GET',
            '/Partitions/id/$/GetReplicas/replicaId',
            ['api-version=6.0'])
        self.validate_command(  # list
            'replica list --continuation-token=ct --partition-id=id',
            'GET',
            '/Partitions/id/$/GetReplicas',
            ['api-version=6.0', 'ContinuationToken=ct'])
        self.validate_command(  # remove
            'replica remove --node-name=nodeName --partition-id=id --replica-id=replicaId --force-remove=true',
            'POST',
            '/Nodes/nodeName/$/GetPartitions/id/$/GetReplicas/replicaId/$/Delete',
            ['api-version=6.0', 'ForceRemove=true'])
        self.validate_command(  # report-health
            'replica report-health --partition-id=id --replica-id=replicaId --service-kind=Stateless --immediate --source-id=ID --health-property=Property ' +
            '--health-state=Warning --ttl=P3Y6M4DT12H30M5S --description=Description ' +
            '--sequence-number=10 --remove-when-expired=true',
            'POST',
            '/Partitions/id/$/GetReplicas/replicaId/$/ReportHealth',
            ['api-version=6.0', 'Immediate=true', 'ServiceKind=Stateless'],
            ('{"SourceId": "ID", '
             '"Property": "Property", '
             '"HealthState": "Warning", '
             '"TimeToLiveInMilliSeconds": "P3Y6M4DT12H30M5S", '
             '"Description": "Description", '
             '"SequenceNumber": "10", '
             '"RemoveWhenExpired": true}'),
            validate_flat_dictionary)
        self.validate_command(  # restart
            'replica restart --node-name=nodeName --partition-id=id --replica-id=replicaId',
            'POST',
            '/Nodes/nodeName/$/GetPartitions/id/$/GetReplicas/replicaId/$/Restart',
            ['api-version=6.0'])

        # RPM
        self.validate_command(  # approve-force
            'rpm approve-force --task-id=id --version=version',
            'POST',
            '/$/ForceApproveRepairTask',
            ['api-version=6.0'],
            '{"TaskId":"id", "Version":"version"}',
            validate_flat_dictionary)
        self.validate_command(  # delete
            'rpm delete --task-id=id --version=version',
            'POST',
            '/$/DeleteRepairTask',
            ['api-version=6.0'],
            '{"TaskId":"id", "Version":"version"}',
            validate_flat_dictionary)
        self.validate_command(  # list
            'rpm list --executor-filter=ex --state-filter=2 --task-id-filter=task',
            'GET',
            '/$/GetRepairTaskList',
            ['api-version=6.0', 'TaskIdFilter=task', 'StateFilter=2', 'ExecutorFilter=ex'])

        # Store
        self.validate_command(  # delete
            'store delete --content-path=path',
            'DELETE',
            '/ImageStore/path',
            ['api-version=6.0'])
        self.validate_command(  # root-info
            'store root-info',
            'GET',
            '/ImageStore',
            ['api-version=6.0'])
        self.validate_command(  # stat
            'store stat --content-path=path',
            'GET',
            '/ImageStore/path',
            ['api-version=6.2'])

        # Service
        # create and update tests not added for now
        self.validate_command(  # app-name
            'service app-name --service-id=some~service~path',
            'GET',
            '/Services/some~service~path/$/GetApplicationName',
            ['api-version=6.0'])
        self.validate_command(  # delete
            'service delete --service-id=some~service~path --force-remove=true',
            'POST',
            '/Services/some~service~path/$/Delete',
            ['api-version=6.0', 'ForceRemove=true'])
        self.validate_command(  # description
            'service description --service-id=some~service~path',
            'GET',
            '/Services/some~service~path/$/GetDescription',
            ['api-version=6.0'])
        self.validate_command(  # health
            'service health --service-id=some~service~path --events-health-state-filter=2 --exclude-health-statistics --partitions-health-state-filter=2',
            'GET',
            '/Services/some~service~path/$/GetHealth',
            ['api-version=6.0', 'EventsHealthStateFilter=2', 'PartitionsHealthStateFilter=2', 'ExcludeHealthStatistics=true'])
        self.validate_command(  # info
            'service info --service-id=some~service~path --application-id=app~id',
            'GET',
            '/Applications/app~id/$/GetServices/some~service~path',
            ['api-version=6.0'])
        self.validate_command(  # list
            'service list --service-type-name=type --application-id=app~id --continuation-token=ct',
            'GET',
            '/Applications/app~id/$/GetServices',
            ['api-version=6.0', 'ServiceTypeName=type', 'ContinuationToken=ct'])
        self.validate_command(  # resolve
            'service resolve --service-id=some~service~path --partition-key-type=3 --partition-key-value=part --previous-rsp-version=version',
            'GET',
            '/Services/some~service~path/$/ResolvePartition',
            ['api-version=6.0', 'PartitionKeyType=3', 'PartitionKeyValue=part', 'PreviousRspVersion=version'])
        self.validate_command(  # report-health
            'service report-health --service-id=service~id --immediate --source-id=ID --health-property=Property ' +
            '--health-state=Warning --ttl=P3Y6M4DT12H30M5S --description=Description ' +
            '--sequence-number=10 --remove-when-expired=true',
            'POST',
            '/Services/service~id/$/ReportHealth',
            ['api-version=6.0', 'Immediate=true'],
            ('{"SourceId": "ID", '
             '"Property": "Property", '
             '"HealthState": "Warning", '
             '"TimeToLiveInMilliSeconds": "P3Y6M4DT12H30M5S", '
             '"Description": "Description", '
             '"SequenceNumber": "10", '
             '"RemoveWhenExpired": true}'),
            validate_flat_dictionary)

        # Chaos commands:
        self.validate_command(  # set chaos schedule
            'chaos schedule set ' +
            '--version 0 --start-date-utc 2016-01-01T00:00:00.000Z ' +
            '--expiry-date-utc 2038-01-01T00:00:00.000Z ' +
            '--chaos-parameters-dictionary [{' +
            '\\\"Key\\\":\\\"adhoc\\\",\\\"Value\\\":{' +
            '\\\"MaxConcurrentFaults\\\":3,\\\"EnableMoveReplicaFaults\\\":true,' +
            '\\\"ChaosTargetFilter\\\":{\\\"NodeTypeInclusionList\\\":[' +
            '\\\"N0010Ref\\\",\\\"N0020Ref\\\",\\\"N0030Ref\\\",\\\"N0040Ref\\\",' +
            '\\\"N0050Ref\\\"]},\\\"MaxClusterStabilizationTimeoutInSeconds\\\":60,' +
            '\\\"WaitTimeBetweenIterationsInSeconds\\\":15,\\\"WaitTimeBetweenFaultsInSeconds\\\":30,' +
            '\\\"TimeToRunInSeconds\\\":\\\"600\\\",\\\"Context\\\":{\\\"Map\\\":{' +
            '\\\"test\\\":\\\"value\\\"}},\\\"ClusterHealthPolicy\\\":{' +
            '\\\"MaxPercentUnhealthyNodes\\\":0,\\\"ConsiderWarningAsError\\\":true,' +
            '\\\"MaxPercentUnhealthyApplications\\\":0}}}] ' +
            '--jobs [{\\\"ChaosParameters\\\":\\\"adhoc\\\",\\\"Days\\\":{' +
            '\\\"Sunday\\\":true,\\\"Monday\\\":true,\\\"Tuesday\\\":true,' +
            '\\\"Wednesday\\\":true,\\\"Thursday\\\":true,\\\"Friday\\\":true,' +
            '\\\"Saturday\\\":true},\\\"Times\\\":[{\\\"StartTime\\\":{' +
            '\\\"Hour\\\":0,\\\"Minute\\\":0},\\\"EndTime\\\":{' +
            '\\\"Hour\\\":23,\\\"Minute\\\":59}}]}]',
            'POST',
            '/Tools/Chaos/Schedule',
            ['api-version=6.2'])

        self.validate_command(  # get chaos schedule
            'chaos schedule get',
            'GET',
            '/Tools/Chaos/Schedule',
            ['api-version=6.2'])

        self.validate_command(  # stop chaos
            'chaos stop',
            'POST',
            '/Tools/Chaos/$/Stop',
            ['api-version=6.0'])

        self.validate_command(  # get chaos events
            'chaos events',
            'GET',
            '/Tools/Chaos/Events',
            ['api-version=6.2'])

        self.validate_command(  # get chaos
            'chaos get',
            'GET',
            '/Tools/Chaos',
            ['api-version=6.2'])

        # Cluster Upgrade Commands:
        self.validate_command(  # config-upgrade
            'sa-cluster config-upgrade --cluster-config YOUR_CLUSTER_CONFIG ' +
            '--application-health-policies {' +
            '\\\"fabric:/System\\\":{' +
            '\\\"ConsiderWarningAsError\\\":true' +
            '}}',
            'POST',
            '/$/StartClusterConfigurationUpgrade',
            ['api-version=6.0', 'timeout=60'],
            ('{"ClusterConfig": "YOUR_CLUSTER_CONFIG", '
             '"HealthCheckRetryTimeout": "P0D", '
             '"HealthCheckWaitDurationInSeconds": "P0D", '
             '"HealthCheckStableDurationInSeconds": "P0D", '
             '"UpgradeDomainTimeoutInSeconds": "P0D", '
             '"UpgradeTimeoutInSeconds": "P0D", '
             '"MaxPercentUnhealthyApplications": 0, '
             '"MaxPercentUnhealthyNodes": 0, '
             '"MaxPercentDeltaUnhealthyNodes": 0, '
             '"MaxPercentUpgradeDomainDeltaUnhealthyNodes": 0, '
             '"ApplicationHealthPolicies": {"ApplicationHealthPolicyMap": [{"Key": "fabric:/System", "Value": {"ConsiderWarningAsError": true}}]}}'),
            validate_flat_dictionary)

        # Events commands:
        self.validate_command(
            'events cluster-list --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Cluster/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events all-nodes-list --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Nodes/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events node-list --node-name=somenode --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Nodes/somenode/$/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events all-applications-list --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Applications/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events application-list --application-id=someappid --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Applications/someappid/$/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events all-services-list --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Services/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events service-list --service-id=someserviceid --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Services/someserviceid/$/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events all-partitions-list --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Partitions/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events partition-list --partition-id=somepartitionid --start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Partitions/somepartitionid/$/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events partition-all-replicas-list --partition-id=somepartitionid ' +
            '--start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Partitions/somepartitionid/$/Replicas/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

        self.validate_command(
            'events partition-replica-list --partition-id=somepartitionid --replica-id=somereplicaid ' +
            '--start-time-utc=2018-01-01T00:00:00Z --end-time-utc=2018-02-01T00:00:00Z',
            'GET',
            '/EventsStore/Partitions/somepartitionid/$/Replicas/somereplicaid/$/Events',
            ['api-version=6.4',
             'StartTimeUtc=' + quote('2018-01-01T00:00:00Z'),
             'EndTimeUtc=' + quote('2018-02-01T00:00:00Z')]
        )

# SIG # Begin Windows Authenticode signature block
# MIIoLQYJKoZIhvcNAQcCoIIoHjCCKBoCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQse8BENmB6EqSR2hd
# JGAGggIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAnIomUPd+4lsSy
# 1HMavKY989Hoq0Ab1wAuU3E77Z0Yy6CCDXYwggX0MIID3KADAgECAhMzAAADrzBA
# DkyjTQVBAAAAAAOvMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMxMTE2MTkwOTAwWhcNMjQxMTE0MTkwOTAwWjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQDOS8s1ra6f0YGtg0OhEaQa/t3Q+q1MEHhWJhqQVuO5amYXQpy8MDPNoJYk+FWA
# hePP5LxwcSge5aen+f5Q6WNPd6EDxGzotvVpNi5ve0H97S3F7C/axDfKxyNh21MG
# 0W8Sb0vxi/vorcLHOL9i+t2D6yvvDzLlEefUCbQV/zGCBjXGlYJcUj6RAzXyeNAN
# xSpKXAGd7Fh+ocGHPPphcD9LQTOJgG7Y7aYztHqBLJiQQ4eAgZNU4ac6+8LnEGAL
# go1ydC5BJEuJQjYKbNTy959HrKSu7LO3Ws0w8jw6pYdC1IMpdTkk2puTgY2PDNzB
# tLM4evG7FYer3WX+8t1UMYNTAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQURxxxNPIEPGSO8kqz+bgCAQWGXsEw
# RQYDVR0RBD4wPKQ6MDgxHjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEW
# MBQGA1UEBRMNMjMwMDEyKzUwMTgyNjAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzci
# tW2oynUClTBUBgNVHR8ETTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5j
# b20vcGtpb3BzL2NybC9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEG
# CCsGAQUFBwEBBFUwUzBRBggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQu
# Y29tL3BraW9wcy9jZXJ0cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0
# MAwGA1UdEwEB/wQCMAAwDQYJKoZIhvcNAQELBQADggIBAISxFt/zR2frTFPB45Yd
# mhZpB2nNJoOoi+qlgcTlnO4QwlYN1w/vYwbDy/oFJolD5r6FMJd0RGcgEM8q9TgQ
# 2OC7gQEmhweVJ7yuKJlQBH7P7Pg5RiqgV3cSonJ+OM4kFHbP3gPLiyzssSQdRuPY
# 1mIWoGg9i7Y4ZC8ST7WhpSyc0pns2XsUe1XsIjaUcGu7zd7gg97eCUiLRdVklPmp
# XobH9CEAWakRUGNICYN2AgjhRTC4j3KJfqMkU04R6Toyh4/Toswm1uoDcGr5laYn
# TfcX3u5WnJqJLhuPe8Uj9kGAOcyo0O1mNwDa+LhFEzB6CB32+wfJMumfr6degvLT
# e8x55urQLeTjimBQgS49BSUkhFN7ois3cZyNpnrMca5AZaC7pLI72vuqSsSlLalG
# OcZmPHZGYJqZ0BacN274OZ80Q8B11iNokns9Od348bMb5Z4fihxaBWebl8kWEi2O
# PvQImOAeq3nt7UWJBzJYLAGEpfasaA3ZQgIcEXdD+uwo6ymMzDY6UamFOfYqYWXk
# ntxDGu7ngD2ugKUuccYKJJRiiz+LAUcj90BVcSHRLQop9N8zoALr/1sJuwPrVAtx
# HNEgSW+AKBqIxYWM4Ev32l6agSUAezLMbq5f3d8x9qzT031jMDT+sUAoCw0M5wVt
# CUQcqINPuYjbS1WgJyZIiEkBMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkq
# hkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5
# IDIwMTEwHhcNMTEwNzA4MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQG
# EwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwG
# A1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQg
# Q29kZSBTaWduaW5nIFBDQSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIIC
# CgKCAgEAq/D6chAcLq3YbqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03
# a8YS2AvwOMKZBrDIOdUBFDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akr
# rnoJr9eWWcpgGgXpZnboMlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0Rrrg
# OGSsbmQ1eKagYw8t00CT+OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy
# 4BI6t0le2O3tQ5GD2Xuye4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9
# sbKvkjh+0p2ALPVOVpEhNSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAh
# dCVfGCi2zCcoOCWYOUo2z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8k
# A/DRelsv1SPjcF0PUUZ3s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTB
# w3J64HLnJN+/RpnF78IcV9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmn
# Eyimp31ngOaKYnhfsi+E11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90
# lfdu+HggWCwTXWCVmj5PM4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0w
# ggHpMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2o
# ynUClTAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYD
# VR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBa
# BgNVHR8EUzBRME+gTaBLhklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2Ny
# bC9wcm9kdWN0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsG
# AQUFBwEBBFIwUDBOBggrBgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraS9jZXJ0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNV
# HSAEgZcwgZQwgZEGCSsGAQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3
# dy5taWNyb3NvZnQuY29tL3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsG
# AQUFBwICMDQeMiAdAEwAZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABl
# AG0AZQBuAHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKb
# C5YR4WOSmUKWfdJ5DJDBZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11l
# hJB9i0ZQVdgMknzSGksc8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6
# I/MTfaaQdION9MsmAkYqwooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0
# wI/zRive/DvQvTXvbiWu5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560
# STkKxgrCxq2u5bLZ2xWIUUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQam
# ASooPoI/E01mC8CzTfXhj38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGa
# J+HNpZfQ7l1jQeNbB5yHPgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ah
# XJbYANahRr1Z85elCUtIEJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA
# 9Z74v2u3S5fi63V4GuzqN5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33Vt
# Y5E90Z1WTk+/gFcioXgRMiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr
# /Xmfwb1tbWrJUnMTDXpQzTGCGg0wghoJAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBIDIwMTECEzMAAAOvMEAOTKNNBUEAAAAAA68wDQYJYIZIAWUDBAIB
# BQCgga4wGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEO
# MAwGCisGAQQBgjcCARUwLwYJKoZIhvcNAQkEMSIEIDMWB+MYo80jmE3CKtolS3Oz
# fPOJ2McfbIqlt8yYCIPcMEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8A
# cwBvAGYAdKEagBhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEB
# BQAEggEAfGVrKFyM/5BDznRSKOjT2ALUaboGTlsmJ66xIwL8kcR/MoTQ7mnqq3R+
# 5U9V21rSl9oiYlKOI5tUqLJXcDUC6CxFyEsJqznHxq8i0fTvO+rY1S0jXk1KpJvk
# fgHgGlAohE0RjtOl6+Z8KmzFDX6i9lcHdXzKFBaRqq//K/yJYUlXj+1FBmnJRMFp
# MZEorOCGQwtyyCwxEsm2bp6YeGiCwg/bTJJSQIxPsp5VritEmu95Jqy/SWgbR6UL
# KOAbRvmY31+cp3lUaHNhyL3egByA4cAfRkllV1YJqAnnymKHoYE86HelrhmPDqP3
# His3Rx401GblR15sdCz8xDW4fL3yF6GCF5cwgheTBgorBgEEAYI3AwMBMYIXgzCC
# F38GCSqGSIb3DQEHAqCCF3AwghdsAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFSBgsq
# hkiG9w0BCRABBKCCAUEEggE9MIIBOQIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFl
# AwQCAQUABCD3WNzR4xbUzJzM08ZAzmU5E2LbNU6BLyHtA0F++wllNgIGZkY1Cc+Y
# GBMyMDI0MDYyMDE1MDAwNC4xMTdaMASAAgH0oIHRpIHOMIHLMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1l
# cmljYSBPcGVyYXRpb25zMScwJQYDVQQLEx5uU2hpZWxkIFRTUyBFU046MzcwMy0w
# NUUwLUQ5NDcxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2Wg
# ghHtMIIHIDCCBQigAwIBAgITMwAAAeqaJHLVWT9hYwABAAAB6jANBgkqhkiG9w0B
# AQsFADB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UE
# BxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYD
# VQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDAeFw0yMzEyMDYxODQ1
# MzBaFw0yNTAzMDUxODQ1MzBaMIHLMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MScwJQYDVQQLEx5uU2hpZWxkIFRTUyBFU046MzcwMy0wNUUwLUQ5NDcxJTAjBgNV
# BAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2UwggIiMA0GCSqGSIb3DQEB
# AQUAA4ICDwAwggIKAoICAQC1C1/xSD8gB9X7Ludoo2rWb2ksqaF65QtJkbQpmsc6
# G4bg5MOv6WP/uJ4XOJvKX/c1t0ej4oWBqdGD6VbjXX4T0KfylTulrzKtgxnxZh7q
# 1uD0Dy/w5G0DJDPb6oxQrz6vMV2Z3y9ZxjfZqBnDfqGon/4VDHnZhdas22svSC5G
# HywsQ2J90MM7L4ecY8TnLI85kXXTVESb09txL2tHMYrB+KHCy08ds36an7IcOGfR
# mhHbFoPa5om9YGpVKS8xeT7EAwW7WbXL/lo5p9KRRIjAlsBBHD1TdGBucrGC3TQX
# STp9s7DjkvvNFuUa0BKsz6UiCLxJGQSZhd2iOJTEfJ1fxYk2nY6SCKsV+VmtV5ai
# PzY/sWoFY542+zzrAPr4elrvr9uB6ci/Kci//EOERZEUTBPXME/ia+t8jrT2y3ug
# 15MSCVuhOsNrmuZFwaRCrRED0yz4V9wlMTGHIJW55iNM3HPVJJ19vOSvrCP9lsEc
# EwWZIQ1FCyPOnkM1fs7880dahAa5UmPqMk5WEKxzDPVp081X5RQ6HGVUz6ZdgQ0j
# cT59EG+CKDPRD6mx8ovzIpS/r/wEHPKt5kOhYrjyQHXc9KHKTWfXpAVj1Syqt5X4
# nr+Mpeubv+N/PjQEPr0iYJDjSzJrqILhBs5pytb6vyR8HUVMp+mAA4rXjOw42vkH
# fQIDAQABo4IBSTCCAUUwHQYDVR0OBBYEFCuBRSWiUebpF0BU1MTIcosFblleMB8G
# A1UdIwQYMBaAFJ+nFV0AXmJdg/Tl0mWnG1M1GelyMF8GA1UdHwRYMFYwVKBSoFCG
# Tmh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvY3JsL01pY3Jvc29mdCUy
# MFRpbWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNybDBsBggrBgEFBQcBAQRgMF4w
# XAYIKwYBBQUHMAKGUGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvY2Vy
# dHMvTWljcm9zb2Z0JTIwVGltZS1TdGFtcCUyMFBDQSUyMDIwMTAoMSkuY3J0MAwG
# A1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYIKwYBBQUHAwgwDgYDVR0PAQH/BAQD
# AgeAMA0GCSqGSIb3DQEBCwUAA4ICAQAog61WXj9+/nxVbX3G37KgvyoNAnuu2w3H
# oWZj3H0YCeQ3b9KSZThVThW4iFcHrKnhFMBbXJX4uQI53kOWSaWCaV3xCznpRt3c
# 4/gSn3dvO/1GP3MJkpJfgo56CgS9zLOiP31kfmpUdPqekZb4ivMR6LoPb5HNlq0W
# bBpzFbtsTjNrTyfqqcqAwc6r99Df2UQTqDa0vzwpA8CxiAg2KlbPyMwBOPcr9hJT
# 8sGpX/ZhLDh11dZcbUAzXHo1RJorSSftVa9hLWnzxGzEGafPUwLmoETihOGLqIQl
# Cpvr94Hiak0Gq0wY6lduUQjk/lxZ4EzAw/cGMek8J3QdiNS8u9ujYh1B7NLr6t3I
# glfScDV3bdVWet1itTUoKVRLIivRDwAT7dRH13Cq32j2JG5BYu/XitRE8cdzaJmD
# VBzYhlPl9QXvC+6qR8I6NIN/9914bTq/S4g6FF4f1dixUxE4qlfUPMixGr0Ft4/S
# 0P4fwmhs+WHRn62PB4j3zCHixKJCsRn9IR3ExBQKQdMi5auiqB6xQBADUf+F7hSK
# ZfbA8sFSFreLSqhvj+qUQF84NcxuaxpbJWVpsO18IL4Qbt45Cz/QMa7EmMGNn7a8
# MM3uTQOlQy0u6c/jq111i1JqMjayTceQZNMBMM5EMc5Dr5m3T4bDj9WTNLgP8SFe
# 3EqTaWVMOTCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkAAAAAABUwDQYJKoZI
# hvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# MjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0eSAy
# MDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVowfDELMAkGA1UEBhMC
# VVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNV
# BAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRp
# bWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoIC
# AQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX9gF/bErg4r25Phdg
# M/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1qUoNEt6aORmsHFPPF
# dvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8dq6z2Nr41JmTamDu6
# GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byNpOORj7I5LFGc6XBp
# Dco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2krnopN6zL64NF50Zu
# yjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4dPf0gz3N9QZpGdc3E
# XzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgSUei/BQOj0XOmTTd0
# lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8QmguEOqEUUbi0b1q
# GFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6CmgyFdXzB0kZSU2LlQ
# +QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzFER1y7435UsSFF5PA
# PBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQIDAQABo4IB3TCCAdkw
# EgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQUKqdS/mTEmr6CkTxG
# NSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1GelyMFwGA1UdIARV
# MFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0dHA6Ly93d3cubWlj
# cm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0bTATBgNVHSUEDDAK
# BggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMC
# AYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbLj+iiXGJo0T2UkFvX
# zpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20v
# cGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0wNi0yMy5jcmwwWgYI
# KwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1pY3Jvc29mdC5j
# b20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIzLmNydDANBgkqhkiG
# 9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwUtj5OR2R4sQaTlz0x
# M7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN3Zi6th542DYunKmC
# VgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU5HhTdSRXud2f8449
# xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5KYnDvBewVIVCs/wM
# nosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGyqVvfSaN0DLzskYDS
# PeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB62FD+CljdQDzHVG2d
# Y3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltEAY5aGZFrDZ+kKNxn
# GSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFpAUR+fKFhbHP+Crvs
# QWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcdFYmNcP7ntdAoGokL
# jzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRbatGePu1+oDEzfbzL
# 6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQdVTNYs6FwZvKhggNQ
# MIICOAIBATCB+aGB0aSBzjCByzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2EgT3BlcmF0aW9uczEn
# MCUGA1UECxMeblNoaWVsZCBUU1MgRVNOOjM3MDMtMDVFMC1EOTQ3MSUwIwYDVQQD
# ExxNaWNyb3NvZnQgVGltZS1TdGFtcCBTZXJ2aWNloiMKAQEwBwYFKw4DAhoDFQCJ
# 2x7cQfjpRskJ8UGIctOCkmEkj6CBgzCBgKR+MHwxCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1w
# IFBDQSAyMDEwMA0GCSqGSIb3DQEBCwUAAgUA6h4skDAiGA8yMDI0MDYyMDA0MjAz
# MloYDzIwMjQwNjIxMDQyMDMyWjB3MD0GCisGAQQBhFkKBAExLzAtMAoCBQDqHiyQ
# AgEAMAoCAQACAgWPAgH/MAcCAQACAhNcMAoCBQDqH34QAgEAMDYGCisGAQQBhFkK
# BAIxKDAmMAwGCisGAQQBhFkKAwKgCjAIAgEAAgMHoSChCjAIAgEAAgMBhqAwDQYJ
# KoZIhvcNAQELBQADggEBAEvdeK316e8b+Kha3MgQkjCTkjujmWyi51pyE9O+yZlF
# +k3BpjkvDOwDi3OT7WmKlNAxC/c0CCbcsAyDf2pqvdr9er8jI8BL0IJpW48Ra4gw
# KYYkORXCpJwSBJk45ykeN0fSslBs/eaQ281GYEfEnyGM8golpG1p1zNhv4SniPhi
# PCoyCjWSXOjfU6E8S9vO8BVBwOQ1EW3Q3H9xfh9L1CxeFkwgZYTR+OpQCKBuEsXp
# +6x6u27sWL3LpMRR3rdI0Sa/THCyE4EvN5nEClep4mh18Eg3YOYCZ74NJ2uJ8x9R
# hZRMprwwybIkoStJAcWE3zpoq9BtpoRGjA9O+TBw0fgxggQNMIIECQIBATCBkzB8
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1N
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAeqaJHLVWT9hYwABAAAB
# 6jANBglghkgBZQMEAgEFAKCCAUowGgYJKoZIhvcNAQkDMQ0GCyqGSIb3DQEJEAEE
# MC8GCSqGSIb3DQEJBDEiBCAZkne5Psxkdna3mKrjJbHGc+9bTmq9BltDpsHVGaYw
# MTCB+gYLKoZIhvcNAQkQAi8xgeowgecwgeQwgb0EICmPodXjZDR4iwg0ltLANXBh
# 5G1uKqKIvq8sjKekuGZ4MIGYMIGApH4wfDELMAkGA1UEBhMCVVMxEzARBgNVBAgT
# Cldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29m
# dCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENB
# IDIwMTACEzMAAAHqmiRy1Vk/YWMAAQAAAeowIgQg/90bK9lyqyYP7/lrv6fwIU1n
# G5deX8tkv+O5WWrvR5MwDQYJKoZIhvcNAQELBQAEggIAbK5RdvNuf8fVl0LHWZSA
# ld69pO7MdMdjn9V5RmWzksMqNZ6bvqb3LHh8IbPgOSjXdQuf9VRrq7xV2XzSutdi
# EwmyMdzqNCUHPSlGLQY9pcHDxjmUgyQ3/zVeNlkRGF2IC08ETwwfGppt4Z4fkDOo
# P08R39UV5nnOShCkIYncgcQ7Tb5kJUNmhlxGh9RUtUUMuTBD8ldZzUK8nFnI0IIT
# suY9x/WXG/CswckHBYnj7sQASXeYFgmTQcvbG/aOWebVyL5PrYGouwvlZzAaew4x
# QagmD3u0PzaZf152usIMnCNFgFaXhraUxBxxD2MqgSInuTmgt7iQamo2nWKJU53M
# U6HJeq7CkDnc0PXk4OiolTr1wjiWY+P1g/RA0/gBaI1MD4eJAupbeOnLRcZ0R5xu
# J9pWf1JtBpWLixx+jOe8HaolOCKTfQ+knKd7ZUtwppsgi6cjYxhZfr+Wq7aKvzrx
# tZ+eArXdtKBkavivPoVMctgweAkYO1wdvVa2PiFGWaAcak87HdOQ94Q18nYoKXmH
# 1pzTXzIG7isAQvc0KzOHd/gjkkAuvhamsC9BJN2qSzFiQEu9BTmLHYNGXi9WqRuk
# U+1la8refRLLJ41dVcEpLHvhM4Zt6kDlc6amTmxh7UdzVvnkOjJSu9+5InlHZdBp
# JLUNBOEeFDrQbtZew/MvOIU=
# SIG # End Windows Authenticode signature block